# -*- coding: utf8 -*-
#
#  Copyright (c) 2017 unfoldingWord
#  http://creativecommons.org/licenses/MIT/
#  See LICENSE file for details.
#
#  Contributors:
#  Richard Mahn <richard_mahn@wycliffeassociates.org>
#  Joel Lonbeck <joel@neutrinographics.com>


from __future__ import unicode_literals, print_function
import os
import json
import boto3
import botocore
from boto3.session import Session
from mimetypes import MimeTypes

class S3Handler(object):
    def __init__(self, bucket_name=None, aws_access_key_id=None, aws_secret_access_key=None, aws_region_name='us-west-2'):
        self.bucket_name = bucket_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_region_name = aws_region_name
        self.bucket = None
        self.client = None
        self.resource = None
        self.setup_resources()

    def setup_resources(self):
        if self.aws_access_key_id and self.aws_secret_access_key:
            session = Session(aws_access_key_id=self.aws_access_key_id,
                              aws_secret_access_key=self.aws_secret_access_key,
                              region_name=self.aws_region_name)
            self.resource = session.resource('s3')
            self.client = session.client('s3')
        else:
            self.resource = boto3.resource('s3')
            self.client = boto3.client('s3')

        self.bucket_name = self.bucket_name
        self.bucket = None
        if self.bucket_name:
            self.bucket = self.resource.Bucket(self.bucket_name)

    def download_file(self, key, local_file):
        """
        Download file from S3 bucket. Similar to s3.download_file except that does
        not play nicely with moto, this however, does.
        :param string key: object to download
        :param string local_file: file to download to
        """
        body = self.resource.Object(bucket_name=self.bucket_name, key=key).get()['Body']
        with open(local_file, 'wb') as f:
            for chunk in iter(lambda: body.read(1024), b''):
                f.write(chunk)

    # Downloads all the files in S3 that have a prefix of `key_prefix` from `bucket` to the `local` directory
    def download_dir(self, key_prefix, local):
        paginator = self.client.get_paginator('list_objects')
        for result in paginator.paginate(Bucket=self.bucket_name, Delimiter='/', Prefix=key_prefix):
            if result.get('CommonPrefixes') is not None:
                for subdir in result.get('CommonPrefixes'):
                    self.download_dir(subdir.get('Prefix'), local)
            if result.get('Contents') is not None:
                for file in result.get('Contents'):
                    local_file = os.path.join(local, file.get('Key'))
                    if local_file.endswith('/'):
                        pass
                    else:
                        if not os.path.exists(os.path.dirname(local_file)):
                            os.makedirs(os.path.dirname(local_file))
                        self.download_file(file.get('Key'), local_file)

    def key_exists(self, key, bucket_name=None):
        if not bucket_name:
            bucket = self.bucket
        else:
            bucket = self.resource.Bucket(bucket_name)

        try:
            bucket.Object(key=key).load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                exists = False
            else:
                raise
        else:
            exists = True

        return exists

    def copy(self, from_key, from_bucket=None, to_key=None, catch_exception=True):
        if not to_key:
            to_key = from_key
        if not from_bucket:
            from_bucket = self.bucket_name

        if catch_exception:
            try:
                return self.resource.Object(bucket_name=self.bucket_name, key=to_key).copy_from(
                    CopySource='{0}/{1}'.format(from_bucket, from_key))
            except:
                return False
        else:
            return self.resource.Object(bucket_name=self.bucket_name, key=to_key).copy_from(
                CopySource='{0}/{1}'.format(from_bucket, from_key))

    def replace(self, key, catch_exception=True):
        if catch_exception:
            try:
                return self.resource.Object(bucket_name=self.bucket_name, key=key).copy_from(
                    CopySource='{0}/{1}'.format(self.bucket_name, key), MetadataDirective='REPLACE')
            except:
                return False
        else:
            return self.resource.Object(bucket_name=self.bucket_name, key=key).copy_from(
                CopySource='{0}/{1}'.format(self.bucket_name, key), MetadataDirective='REPLACE')

    def upload_file(self, path, key, cache_time=600):
        """
        Upload file to S3 storage. Similar to the s3.upload_file, however, that
        does not work nicely with moto, whereas this function does.
        :param string path: file to upload
        :param string key: name of the object in the bucket
        """
        with open(path, 'rb') as f:
            binary = f.read()
        self.bucket.put_object(
            Key=key,
            Body=binary,
            ContentType=self._get_mime_type(path),
            CacheControl='max-age={0}'.format(cache_time)
        )

    def get_object(self, key):
        return self.resource.Object(bucket_name=self.bucket_name, key=key)

    def redirect(self, key, location):
        self.bucket.put_object(Key=key, WebsiteRedirectLocation=location, CacheControl='max-age=0')

    def get_file_contents(self, key, catch_exception=True):
        if catch_exception:
            try:
                return self.get_object(key).get()['Body'].read().decode('utf-8')
            except:
                return None
        else:
            return self.get_object(key).get()['Body'].read().decode('utf-8')

    def get_json(self, key, catch_exception = True):
        if catch_exception:
            try:
                return json.loads(self.get_file_contents(key))
            except:
                return {}
        else:
            return json.loads(self.get_file_contents(key, catch_exception))

    def get_objects(self, prefix=None, suffix=None):
        filtered = []
        objects = self.bucket.objects.filter(Prefix=prefix)
        if objects:
            if suffix:
                for obj in objects:
                    if obj.key.endswith(suffix):
                        filtered.append(obj)
            else:
                filtered = objects
        return filtered

    def put_contents(self, key, body, catch_exception=True):
        if catch_exception:
            try:
                return self.get_object(key).put(Body=body)
            except:
                return None
        else:
            return self.get_object(key).put(Body=body)

    def delete_file(self, key, catch_exception=True):
        if catch_exception:
            try:
                return self.resource.Object(bucket_name=self.bucket_name, key=key).delete()
            except:
                return False
        else:
            return self.resource.Object(bucket_name=self.bucket_name, key=key).delete()

    def create_bucket(self, bucket_name=None, catch_exception=True):
        if not bucket_name:
            bucket_name = self.bucket_name
        if catch_exception:
            try:
                return self.resource.create_bucket(Bucket=bucket_name)
            except:
                return None
        else:
            return self.resource.create_bucket(Bucket=bucket_name)

    @staticmethod
    def _get_mime_type(path):
        mime = MimeTypes()

        mime_type = mime.guess_type(path)[0]
        if not mime_type:
            mime_type = "text/{0}".format(os.path.splitext(path)[1])
        return mime_type


# -*- coding: utf8 -*-
#
#  Copyright (c) 2017 unfoldingWord
#  http://creativecommons.org/licenses/MIT/
#  See LICENSE file for details.
#
#  Contributors:
#  Richard Mahn <richard_mahn@wycliffeassociates.org>
#  Joel Lonbeck <joel@neutrinographics.com>

import boto3

from boto3.session import Session


class SESHandler(object):
    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None, aws_region_name='us-west-2'):
        if aws_access_key_id and aws_secret_access_key:
            session = Session(aws_access_key_id=aws_access_key_id,
                              aws_secret_access_key=aws_secret_access_key,
                              region_name=aws_region_name)
            self.client = session.client('ses')
        else:
            self.client = boto3.client('ses')

    def send_email(self, **kwargs):
        return self.client.send_email(**kwargs)



from setuptools import setup, find_packages

setup(
    name='d43-aws-tools',
    version='1.0.3',
    packages=find_packages(exclude=['tests', 'general_tools']),
    author='unfoldingWord',
    author_email='unfoldingword.org',
    description='Unit test setup file.',
    keywords=[
        'aws tools',
        'unfoldingword',
        'lambda'
    ],
    url='https://github.org/unfoldingWord-dev/python-aws-tools',
    long_description='Unit test setup file',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires=[
        'coveralls==1.1',
        'boto3==1.4.4',
        'moto==0.4.31',
        'mock',  # travis reports syntax error in mock setup.cfg if we give version
    ],
    test_suite='tests'
)



import os
from setuptools import setup, find_packages



def read(f_name):
    """
    Utility function to read the README file.

    Used for the long_description.  It's nice, because now 1) we have a top level
    README file and 2) it's easier to type in the README file than to put a raw
    string in below ...
    """
    here = os.path.abspath(os.path.dirname(__file__))
    return open(os.path.join(here, f_name), 'r').read()

setup(
    name='d43-aws-tools',
    version='1.0.3',
    packages=find_packages(exclude=['tests', 'general_tools']),
    author='unfoldingWord',
    author_email='info@unfoldingword.org',
    description='Classes for accessing AWS APIs',
    license='MIT',
    keywords=[
        'aws tools',
        'unfoldingword',
        'lambda'
    ],
    url='https://github.org/unfoldingWord-dev/python-aws-tools',
    long_description=read('README.rst'),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
    ],
    install_requires=[
        'boto3==1.4.4',
        'moto==1.0.1',
        'mock==2.0.0'
    ]
)



# -*- coding: utf8 -*-
#
#  Copyright (c) 2017 unfoldingWord
#  http://creativecommons.org/licenses/MIT/
#  See LICENSE file for details.
#
#  Contributors:
#  Richard Mahn <richard_mahn@wycliffeassociates.org>
#  Joel Lonbeck <joel@neutrinographics.com>


from __future__ import unicode_literals, print_function
import boto3
import logging
from six import iteritems
from boto3 import Session
from boto3.dynamodb.conditions import Attr, Key


class DynamoDBHandler(object):

    def __init__(self, table_name, aws_access_key_id=None, aws_secret_access_key=None, aws_region_name='us-west-2'):
        self.table_name = table_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_region_name = aws_region_name
        self.resource = None
        self.table = None
        self.logger = logging.getLogger()
        self.setup_resources()

    def setup_resources(self):
        if self.aws_access_key_id and self.aws_secret_access_key:
            session = Session(aws_access_key_id=self.aws_access_key_id,
                              aws_secret_access_key=self.aws_secret_access_key,
                              region_name=self.aws_region_name)
            self.resource = session.resource('dynamodb')
        else:
            self.resource = boto3.resource('dynamodb')
        self.table = self.resource.Table(self.table_name)

    def get_item(self, keys):
        response = self.table.get_item(
            Key=keys
        )
        if 'Item' in response:
            return response['Item']
        else:
            return None

    def insert_item(self, data):
        self.table.put_item(
            Item=data
        )

    def update_item(self, keys, data):
        expressions = []
        names = {}
        values = {}

        for field, value in iteritems(data):
            if field not in keys:
                name = field
                if name.upper() in RESERVED_WORDS:
                    name = '#item_'+name
                    names[name] = field
                expressions.append('{0} = :{1}'.format(name, field))
                values[':{0}'.format(field)] = value

        self.logger.debug('UPDATING WITH:')
        self.logger.debug('SET {0}'.format(', '.join(expressions)))
        self.logger.debug("VALUES:")
        self.logger.debug(values)
        self.logger.debug("NAMES:")
        self.logger.debug(names)

        if names:
            return self.table.update_item(
                Key=keys,
                UpdateExpression='SET {0}'.format(', '.join(expressions)),
                ExpressionAttributeValues=values,
                ExpressionAttributeNames=names
            )
        else:
            return self.table.update_item(
                Key=keys,
                UpdateExpression='SET {0}'.format(', '.join(expressions)),
                ExpressionAttributeValues=values
            )

    def delete_item(self, keys):
        return self.table.delete_item(
            Key=keys
        )

    def get_item_count(self):
        """
        get number of items in table - one caveat is that this value may be off since AWS only updates it every 6 hours
        :return:
        """
        return self.table.item_count

    def query_items(self, query=None, only_fields_with_values=True, queryChunkLimit=-1):
        """
        gets items from database
        :param query:
        :param only_fields_with_values:
        :param queryChunkLimit: not an absolute count, but a threshold where we stop fetching more chunks
                        (if negative then no limit, but will read all chunks)
        :return:
        """
        filter_expression = None
        if query and len(query) >= 1:
            for field, value in iteritems(query):
                value2 = None
                if isinstance(value, dict) and 'condition' in value and 'value' in value:
                    condition_str = value['condition']
                    value = value['value']
                    if condition_str == "between":
                        value, value2 = value["value"], value["value2"]
                else:
                    condition_str = 'eq'

                if not isinstance(value, bool) and not value and only_fields_with_values:
                    continue

                if condition_str == 'eq':
                    condition = Attr(field).eq(value)
                elif condition_str == 'ne':
                    condition = Attr(field).ne(value)
                elif condition_str == 'lt':
                    condition = Attr(field).lt(value)
                elif condition_str == 'lte':
                    condition = Attr(field).lte(value)
                elif condition_str == 'gt':
                    condition = Attr(field).gt(value)
                elif condition_str == 'gte':
                    condition = Attr(field).gte(value)
                elif condition_str == 'begins_with':
                    condition = Attr(field).begins_with(value)
                elif condition_str == 'between':
                    condition = Attr(field).between(value, value2)
                elif condition_str == 'is_in':
                    condition = Attr(field).is_in(value)
                elif condition_str == 'contains':
                    condition = Attr(field).contains(value)
                else:
                    raise Exception('Invalid filter condition: {0}'.format(condition_str))

                if filter_expression is None:
                    filter_expression = condition
                else:
                    filter_expression &= condition

        if filter_expression is not None:
            response = self.table.scan(
                FilterExpression=filter_expression
            )
        else:
            response = self.table.scan()

        if not response or not('Items' in response):
            return None

        # finished if there is no more data to read
        if not('LastEvaluatedKey' in response):
            return response['Items']

        items = response['Items']

        # read chunks until end or threshold is reached
        while 'LastEvaluatedKey' in response:
            if filter_expression is not None:
                response = self.table.scan(
                    FilterExpression=filter_expression,
                    ExclusiveStartKey = response['LastEvaluatedKey']
                )
            else:
                response = self.table.scan(ExclusiveStartKey = response['LastEvaluatedKey'])

            if response and ('Items' in response):
                items += response['Items']

            itemCount = len(items)
            if (queryChunkLimit >= 0) and (itemCount >= queryChunkLimit):
                break

        return items

RESERVED_WORDS = [
    'ABORT',
    'ABSOLUTE',
    'ACTION',
    'ADD',
    'AFTER',
    'AGENT',
    'AGGREGATE',
    'ALL',
    'ALLOCATE',
    'ALTER',
    'ANALYZE',
    'AND',
    'ANY',
    'ARCHIVE',
    'ARE',
    'ARRAY',
    'AS',
    'ASC',
    'ASCII',
    'ASENSITIVE',
    'ASSERTION',
    'ASYMMETRIC',
    'AT',
    'ATOMIC',
    'ATTACH',
    'ATTRIBUTE',
    'AUTH',
    'AUTHORIZATION',
    'AUTHORIZE',
    'AUTO',
    'AVG',
    'BACK',
    'BACKUP',
    'BASE',
    'BATCH',
    'BEFORE',
    'BEGIN',
    'BETWEEN',
    'BIGINT',
    'BINARY',
    'BIT',
    'BLOB',
    'BLOCK',
    'BOOLEAN',
    'BOTH',
    'BREADTH',
    'BUCKET',
    'BULK',
    'BY',
    'BYTE',
    'CALL',
    'CALLED',
    'CALLING',
    'CAPACITY',
    'CASCADE',
    'CASCADED',
    'CASE',
    'CAST',
    'CATALOG',
    'CHAR',
    'CHARACTER',
    'CHECK',
    'CLASS',
    'CLOB',
    'CLOSE',
    'CLUSTER',
    'CLUSTERED',
    'CLUSTERING',
    'CLUSTERS',
    'COALESCE',
    'COLLATE',
    'COLLATION',
    'COLLECTION',
    'COLUMN',
    'COLUMNS',
    'COMBINE',
    'COMMENT',
    'COMMIT',
    'COMPACT',
    'COMPILE',
    'COMPRESS',
    'CONDITION',
    'CONFLICT',
    'CONNECT',
    'CONNECTION',
    'CONSISTENCY',
    'CONSISTENT',
    'CONSTRAINT',
    'CONSTRAINTS',
    'CONSTRUCTOR',
    'CONSUMED',
    'CONTINUE',
    'CONVERT',
    'COPY',
    'CORRESPONDING',
    'COUNT',
    'COUNTER',
    'CREATE',
    'CROSS',
    'CUBE',
    'CURRENT',
    'CURSOR',
    'CYCLE',
    'DATA',
    'DATABASE',
    'DATE',
    'DATETIME',
    'DAY',
    'DEALLOCATE',
    'DEC',
    'DECIMAL',
    'DECLARE',
    'DEFAULT',
    'DEFERRABLE',
    'DEFERRED',
    'DEFINE',
    'DEFINED',
    'DEFINITION',
    'DELETE',
    'DELIMITED',
    'DEPTH',
    'DEREF',
    'DESC',
    'DESCRIBE',
    'DESCRIPTOR',
    'DETACH',
    'DETERMINISTIC',
    'DIAGNOSTICS',
    'DIRECTORIES',
    'DISABLE',
    'DISCONNECT',
    'DISTINCT',
    'DISTRIBUTE',
    'DO',
    'DOMAIN',
    'DOUBLE',
    'DROP',
    'DUMP',
    'DURATION',
    'DYNAMIC',
    'EACH',
    'ELEMENT',
    'ELSE',
    'ELSEIF',
    'EMPTY',
    'ENABLE',
    'END',
    'EQUAL',
    'EQUALS',
    'ERROR',
    'ESCAPE',
    'ESCAPED',
    'EVAL',
    'EVALUATE',
    'EXCEEDED',
    'EXCEPT',
    'EXCEPTION',
    'EXCEPTIONS',
    'EXCLUSIVE',
    'EXEC',
    'EXECUTE',
    'EXISTS',
    'EXIT',
    'EXPLAIN',
    'EXPLODE',
    'EXPORT',
    'EXPRESSION',
    'EXTENDED',
    'EXTERNAL',
    'EXTRACT',
    'FAIL',
    'FALSE',
    'FAMILY',
    'FETCH',
    'FIELDS',
    'FILE',
    'FILTER',
    'FILTERING',
    'FINAL',
    'FINISH',
    'FIRST',
    'FIXED',
    'FLATTERN',
    'FLOAT',
    'FOR',
    'FORCE',
    'FOREIGN',
    'FORMAT',
    'FORWARD',
    'FOUND',
    'FREE',
    'FROM',
    'FULL',
    'FUNCTION',
    'FUNCTIONS',
    'GENERAL',
    'GENERATE',
    'GET',
    'GLOB',
    'GLOBAL',
    'GO',
    'GOTO',
    'GRANT',
    'GREATER',
    'GROUP',
    'GROUPING',
    'HANDLER',
    'HASH',
    'HAVE',
    'HAVING',
    'HEAP',
    'HIDDEN',
    'HOLD',
    'HOUR',
    'IDENTIFIED',
    'IDENTITY',
    'IF',
    'IGNORE',
    'IMMEDIATE',
    'IMPORT',
    'IN',
    'INCLUDING',
    'INCLUSIVE',
    'INCREMENT',
    'INCREMENTAL',
    'INDEX',
    'INDEXED',
    'INDEXES',
    'INDICATOR',
    'INFINITE',
    'INITIALLY',
    'INLINE',
    'INNER',
    'INNTER',
    'INOUT',
    'INPUT',
    'INSENSITIVE',
    'INSERT',
    'INSTEAD',
    'INT',
    'INTEGER',
    'INTERSECT',
    'INTERVAL',
    'INTO',
    'INVALIDATE',
    'IS',
    'ISOLATION',
    'ITEM',
    'ITEMS',
    'ITERATE',
    'JOIN',
    'KEY',
    'KEYS',
    'LAG',
    'LANGUAGE',
    'LARGE',
    'LAST',
    'LATERAL',
    'LEAD',
    'LEADING',
    'LEAVE',
    'LEFT',
    'LENGTH',
    'LESS',
    'LEVEL',
    'LIKE',
    'LIMIT',
    'LIMITED',
    'LINES',
    'LIST',
    'LOAD',
    'LOCAL',
    'LOCALTIME',
    'LOCALTIMESTAMP',
    'LOCATION',
    'LOCATOR',
    'LOCK',
    'LOCKS',
    'LOG',
    'LOGED',
    'LONG',
    'LOOP',
    'LOWER',
    'MAP',
    'MATCH',
    'MATERIALIZED',
    'MAX',
    'MAXLEN',
    'MEMBER',
    'MERGE',
    'METHOD',
    'METRICS',
    'MIN',
    'MINUS',
    'MINUTE',
    'MISSING',
    'MOD',
    'MODE',
    'MODIFIES',
    'MODIFY',
    'MODULE',
    'MONTH',
    'MULTI',
    'MULTISET',
    'NAME',
    'NAMES',
    'NATIONAL',
    'NATURAL',
    'NCHAR',
    'NCLOB',
    'NEW',
    'NEXT',
    'NO',
    'NONE',
    'NOT',
    'NULL',
    'NULLIF',
    'NUMBER',
    'NUMERIC',
    'OBJECT',
    'OF',
    'OFFLINE',
    'OFFSET',
    'OLD',
    'ON',
    'ONLINE',
    'ONLY',
    'OPAQUE',
    'OPEN',
    'OPERATOR',
    'OPTION',
    'OR',
    'ORDER',
    'ORDINALITY',
    'OTHER',
    'OTHERS',
    'OUT',
    'OUTER',
    'OUTPUT',
    'OVER',
    'OVERLAPS',
    'OVERRIDE',
    'OWNER',
    'PAD',
    'PARALLEL',
    'PARAMETER',
    'PARAMETERS',
    'PARTIAL',
    'PARTITION',
    'PARTITIONED',
    'PARTITIONS',
    'PATH',
    'PERCENT',
    'PERCENTILE',
    'PERMISSION',
    'PERMISSIONS',
    'PIPE',
    'PIPELINED',
    'PLAN',
    'POOL',
    'POSITION',
    'PRECISION',
    'PREPARE',
    'PRESERVE',
    'PRIMARY',
    'PRIOR',
    'PRIVATE',
    'PRIVILEGES',
    'PROCEDURE',
    'PROCESSED',
    'PROJECT',
    'PROJECTION',
    'PROPERTY',
    'PROVISIONING',
    'PUBLIC',
    'PUT',
    'QUERY',
    'QUIT',
    'QUORUM',
    'RAISE',
    'RANDOM',
    'RANGE',
    'RANK',
    'RAW',
    'READ',
    'READS',
    'REAL',
    'REBUILD',
    'RECORD',
    'RECURSIVE',
    'REDUCE',
    'REF',
    'REFERENCE',
    'REFERENCES',
    'REFERENCING',
    'REGEXP',
    'REGION',
    'REINDEX',
    'RELATIVE',
    'RELEASE',
    'REMAINDER',
    'RENAME',
    'REPEAT',
    'REPLACE',
    'REQUEST',
    'RESET',
    'RESIGNAL',
    'RESOURCE',
    'RESPONSE',
    'RESTORE',
    'RESTRICT',
    'RESULT',
    'RETURN',
    'RETURNING',
    'RETURNS',
    'REVERSE',
    'REVOKE',
    'RIGHT',
    'ROLE',
    'ROLES',
    'ROLLBACK',
    'ROLLUP',
    'ROUTINE',
    'ROW',
    'ROWS',
    'RULE',
    'RULES',
    'SAMPLE',
    'SATISFIES',
    'SAVE',
    'SAVEPOINT',
    'SCAN',
    'SCHEMA',
    'SCOPE',
    'SCROLL',
    'SEARCH',
    'SECOND',
    'SECTION',
    'SEGMENT',
    'SEGMENTS',
    'SELECT',
    'SELF',
    'SEMI',
    'SENSITIVE',
    'SEPARATE',
    'SEQUENCE',
    'SERIALIZABLE',
    'SESSION',
    'SET',
    'SETS',
    'SHARD',
    'SHARE',
    'SHARED',
    'SHORT',
    'SHOW',
    'SIGNAL',
    'SIMILAR',
    'SIZE',
    'SKEWED',
    'SMALLINT',
    'SNAPSHOT',
    'SOME',
    'SOURCE',
    'SPACE',
    'SPACES',
    'SPARSE',
    'SPECIFIC',
    'SPECIFICTYPE',
    'SPLIT',
    'SQL',
    'SQLCODE',
    'SQLERROR',
    'SQLEXCEPTION',
    'SQLSTATE',
    'SQLWARNING',
    'START',
    'STATE',
    'STATIC',
    'STATUS',
    'STORAGE',
    'STORE',
    'STORED',
    'STREAM',
    'STRING',
    'STRUCT',
    'STYLE',
    'SUB',
    'SUBMULTISET',
    'SUBPARTITION',
    'SUBSTRING',
    'SUBTYPE',
    'SUM',
    'SUPER',
    'SYMMETRIC',
    'SYNONYM',
    'SYSTEM',
    'TABLE',
    'TABLESAMPLE',
    'TEMP',
    'TEMPORARY',
    'TERMINATED',
    'TEXT',
    'THAN',
    'THEN',
    'THROUGHPUT',
    'TIME',
    'TIMESTAMP',
    'TIMEZONE',
    'TINYINT',
    'TO',
    'TOKEN',
    'TOTAL',
    'TOUCH',
    'TRAILING',
    'TRANSACTION',
    'TRANSFORM',
    'TRANSLATE',
    'TRANSLATION',
    'TREAT',
    'TRIGGER',
    'TRIM',
    'TRUE',
    'TRUNCATE',
    'TTL',
    'TUPLE',
    'TYPE',
    'UNDER',
    'UNDO',
    'UNION',
    'UNIQUE',
    'UNIT',
    'UNKNOWN',
    'UNLOGGED',
    'UNNEST',
    'UNPROCESSED',
    'UNSIGNED',
    'UNTIL',
    'UPDATE',
    'UPPER',
    'URL',
    'USAGE',
    'USE',
    'USER',
    'USERS',
    'USING',
    'UUID',
    'VACUUM',
    'VALUE',
    'VALUED',
    'VALUES',
    'VARCHAR',
    'VARIABLE',
    'VARIANCE',
    'VARINT',
    'VARYING',
    'VIEW',
    'VIEWS',
    'VIRTUAL',
    'VOID',
    'WAIT',
    'WHEN',
    'WHENEVER',
    'WHERE',
    'WHILE',
    'WINDOW',
    'WITH',
    'WITHIN',
    'WITHOUT',
    'WORK',
    'WRAPPED',
    'WRITE',
    'YEAR',
    'ZONE'
]
